/*
 * Copyright 2021 Red Hat, Inc. and/or its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.kie.kogito.serialization.process.impl;

import java.io.InputStream;
import java.io.OutputStream;

import org.kie.kogito.serialization.process.MarshallerReaderContext;
import org.kie.kogito.serialization.process.MarshallerWriterContext;
import org.kie.kogito.serialization.process.ProcessInstanceMarshaller;
import org.kie.kogito.serialization.process.ProcessInstanceMarshallerFactory;

public class ProtobufProcessInstanceMarshallerFactory implements ProcessInstanceMarshallerFactory {

    public ProcessInstanceMarshaller newKogitoProcessInstanceMarshaller() {
        return new ProtobufProcessInstanceMarshaller();
    }

    public MarshallerWriterContext newWriterContext(OutputStream baos) {
        return new ProtobufProcessMarshallerWriteContext(baos);
    }

    public MarshallerReaderContext newReaderContext(InputStream bais) {
        return new ProtobufMarshallerReaderContext(bais);
    }
}
